/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.examples.netDesignAlgorithm.fa;

import com.jom.OptimizationProblem;
import com.net2plan.interfaces.networkDesign.IAlgorithm;
import com.net2plan.interfaces.networkDesign.Net2PlanException;
import com.net2plan.interfaces.networkDesign.NetPlan;
import com.net2plan.libraries.CandidatePathList;
import com.net2plan.utils.CollectionUtils;
import com.net2plan.utils.Constants;
import com.net2plan.utils.DoubleUtils;
import com.net2plan.utils.Triple;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Given a network topology, the capacities in the links and the offered traffic, this algorithm obtains the traffic routing that minimizes
 * the worst link utilization using a flow-path formulation. For each demand, the set of admissible paths is composed of the ranking of (at most)
 * "k" shortest paths in km, between the demand end nodes. "k" is a user-defined parameter.  Paths for which its propagation time sums more than
 * "maxPropDelayMs", a user-defined parameter, are considered not admissible. The routing may be constrained to be non-bifurcated
 * setting the user-defined parameter "nonBifurcated" to "true".
 *
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @version 1.1, May 2015
 */
public class FA_minBottleneckUtilization_xp implements IAlgorithm
{
	@Override
	public String executeAlgorithm(NetPlan netPlan, Map<String, String> algorithmParameters, Map<String, String> net2planParameters)
	{
		/* Basic checks */
		int N = netPlan.getNumberOfNodes();
		int E = netPlan.getNumberOfLinks();
		int D = netPlan.getNumberOfDemands();
		if (N == 0 || D == 0 || E == 0) throw new Net2PlanException("This algorithm requires a topology with links, and a demand set");

		/* Initialize some variables */
		double[] u_e = netPlan.getLinkCapacityVector();
		Map<Long, Double> h_d = netPlan.getDemandOfferedTrafficMap();
		int k = Integer.parseInt(algorithmParameters.get("k"));
		boolean nonBifurcatedRouting = Boolean.parseBoolean(algorithmParameters.get("nonBifurcated"));
		double maxPropDelayInMs = Double.parseDouble(algorithmParameters.get("maxPropDelayInMs"));

		/* Remove all routes in current netPlan object */
		netPlan.removeAllRoutingInformation();
		netPlan.setRoutingType(Constants.RoutingType.SOURCE_ROUTING);

		/* Create the candidate path list */
		CandidatePathList cpl = new CandidatePathList(netPlan, netPlan.getLinkLengthInKmMap(), "K", Integer.toString(k), "maxPropDelayInMs", Double.toString(maxPropDelayInMs));
		int P = cpl.getNumberOfPaths();
		double[] h_p = cpl.computePathMaximumCarriedTrafficVector(h_d); /* for each path p, the offered traffic of its associated demand d(p) */

		/* Create the optimization problem object (JOM library) */
		OptimizationProblem op = new OptimizationProblem();

		/* Add the decision variables to the problem */
		op.addDecisionVariable("x_p", nonBifurcatedRouting, new int[] { 1, P }, 0, 1); /* the fraction of traffic of demand d(p) that is carried by p */
		op.addDecisionVariable("rho", false, new int[] { 1, 1 }, 0, 1); /* the worse utilization in the links */

		/* Set some input parameters */
		op.setInputParameter("u_e", u_e, "row");
		op.setInputParameter("h_p", h_p, "row");

		/* Sets the objective function */
		op.setObjectiveFunction("minimize", "rho");

		/* VECTORIAL FORM OF THE CONSTRAINTS  */
		op.setInputParameter("A_dp", cpl.getDemand2PathAssignmentMatrix());
		op.setInputParameter("A_ep", cpl.getLink2PathAssignmentMatrix());
		op.addConstraint("A_dp * x_p' == 1"); /* for each demand, the 100% of the traffic is carried (summing the associated paths) */
		op.addConstraint("A_ep * ((h_p .* x_p)') <= u_e' * rho", "capConstraints"); /* for each link, its utilization is below or equal to rho */

		/* Call the solver to solve the problem */
		String solverName = algorithmParameters.get("solverName");
		String solverLibraryName = algorithmParameters.get("solverLibraryName");
		op.solve(solverName, "solverLibraryName", solverLibraryName);

		/* If an optimal solution was not found, quit */
		if (!op.solutionIsOptimal()) throw new RuntimeException("An optimal solution was not found");

		/* Retrieve the optimum solutions */
		double[] pi_e = op.getMultipliersOfConstraint("capConstraints").to1DArray();
		double[] x_p = op.getPrimalSolution("x_p").to1DArray();
		x_p = DoubleUtils.mult(x_p, h_p);

		/* Update netPlan object: routing */
		netPlan.removeAllRoutingInformation();
		netPlan.setRoutingType(Constants.RoutingType.SOURCE_ROUTING);
		netPlan.addRoutes(cpl, x_p, false);
		
		/* Print the multipliers of the link utilization constraints */
		Set<Long> linkIds = netPlan.getLinkIds();
		Map<Long, Integer> linkId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(linkIds);
		for (long linkId : linkIds)
		{
			int e = linkId2LinearIndex.get(linkId);
			long originNodeId = netPlan.getLinkOriginNode(linkId);
			long destinationNodeId = netPlan.getLinkDestinationNode(linkId);
			
			System.out.println("Link id " + e + "(" + originNodeId + " -> " + destinationNodeId + "): pi = " + pi_e[e] + " , util " + netPlan.getLinkUtilization(linkId));
		}

		return "Ok!: |P| = " + cpl.getNumberOfPaths();
	}

	@Override
	public String getDescription()
	{
		return "Given a network topology, the capacities in the links and the offered traffic, this algorithm obtains the traffic routing that minimizes"
				+ " the worst link utilization using a flow-path formulation. For each demand, the set of admissible paths is composed of the ranking of (at most) "
				+ "\"k\" shortest paths in km, between the demand end nodes. \"k\" is a user-defined parameter.  Paths for which its propagation time sums more than "
				+ " \"maxPropDelayMs\", a user-defined parameter, are considered not admissible. The routing may be constrained to be non-bifurcated "
				+ " setting the user-defined parameter \"nonBifurcated\" to \"true\"";
	}

	@Override
	public List<Triple<String, String, String>> getParameters()
	{
		List<Triple<String, String, String>> algorithmParameters = new LinkedList<Triple<String, String, String>>();
		algorithmParameters.add(Triple.of("k", "5", "Maximum number of admissible paths per demand"));
		algorithmParameters.add(Triple.of("nonBifurcated", "#boolean# false", "True if the routing is constrained to be non-bifurcated"));
		algorithmParameters.add(Triple.of("maxPropDelayInMs", "30", "Maximum propagation delay in miliseconds allowed in a path"));
		algorithmParameters.add(Triple.of("solverName", "glpk", "The solver name to be used by JOM"));
		algorithmParameters.add(Triple.of("solverLibraryName", "", "The solver library full or relative path, to be used by JOM. Leave blank to use JOM default"));
		
		return algorithmParameters;
	}
}
